import logging

from django.apps import AppConfig
from django.db import OperationalError, ProgrammingError
from django.template import TemplateDoesNotExist

from content_blocks.conf import settings

logger = logging.getLogger(__name__)


class ContentBlocksConfig(AppConfig):
    name = "content_blocks"
    verbose_name = "Content blocks"

    def ready(self):
        from content_blocks.models import ContentBlock
        from content_blocks.signals import (  # noqa
            cleanup_media_delete,
            cleanup_media_save,
        )

        if not settings.CONTENT_BLOCKS_DISABLE_UPDATE_CACHE_MODEL_CHOICE:
            from content_blocks.signals import update_cache_model_choice  # noqa

        if "dbtemplates" in settings.INSTALLED_APPS:
            from content_blocks.signals import update_cache_template  # noqa

        if not settings.CONTENT_BLOCKS_DISABLE_CACHE_ON_START:
            try:
                for content_block in ContentBlock.objects.filter(
                    parent__isnull=True,
                    content_block_template__template_filename__isnull=False,
                    draft=False,
                ):
                    try:
                        content_block.render()
                    except TemplateDoesNotExist:  # pragma: no cover
                        logger.error(
                            f"No template found when rendering {content_block}"
                        )

            except (OperationalError, ProgrammingError):  # pragma: no cover
                # Migrate hasn't been run yet.
                pass  # pragma: no cover



"""
Content blocks template tags.
"""
from django import template

from content_blocks.models import ContentBlockCollection

register = template.Library()


@register.inclusion_tag("content_blocks/content_block_collection.html")
def content_block_collection(content_block_collection_slug):
    """
    Render content blocks for a ContentBlockCollection given the collection's slug.
    """
    context = {"slug": content_block_collection_slug}

    try:
        collection = ContentBlockCollection.objects.get(
            slug=content_block_collection_slug
        )
        context.update({"content_block_collection": collection})
    except ContentBlockCollection.DoesNotExist:
        pass

    return context



"""
Content blocks app signals.py
"""
from django.contrib.contenttypes.models import ContentType
from django.db import IntegrityError, OperationalError, ProgrammingError
from django.db.migrations.recorder import MigrationRecorder
from django.db.models.signals import post_delete, post_save, pre_delete, pre_save
from django.dispatch import receiver

from content_blocks.conf import settings
from content_blocks.models import (
    ContentBlock,
    ContentBlockField,
    ContentBlockFields,
    FileField,
    ImageField,
    VideoField,
)

if not settings.CONTENT_BLOCKS_DISABLE_UPDATE_CACHE_MODEL_CHOICE:

    @receiver(post_save, dispatch_uid="update_cache_model_choice_save")
    @receiver(post_delete, dispatch_uid="update_cache_model_choice_delete")
    def update_cache_model_choice(sender, instance, **kwargs):
        """
        Clear the cache for content blocks when related objects are saved. Model choice fields.
        :param instance:
        :param sender:
        :param kwargs:
        :return:
        """
        if kwargs.get("raw", False):
            # Prevent this signal from running during loaddata.
            return

        if sender == MigrationRecorder.Migration:
            # Do not run the signal for MigrationRecorder.Migration otherwise migrations will fail
            return

        try:
            content_block_fields = ContentBlockField.objects.filter(
                model_choice_content_type=ContentType.objects.get_for_model(sender),
                model_choice_object_id=getattr(instance, "id", None),
            ).select_related("content_block")

            for content_block_field in content_block_fields:
                content_block_field.content_block.update_cache()

        except (
            OperationalError,
            ProgrammingError,
            IntegrityError,
            ContentType.DoesNotExist,
        ):  # pragma: no cover
            """
            Content blocks not migrated yet.
            """
            return


if "dbtemplates" in settings.INSTALLED_APPS:
    from dbtemplates.models import Template

    @receiver(post_save, sender=Template, dispatch_uid="update_cache_template_save")
    @receiver(post_delete, sender=Template, dispatch_uid="update_cache_template_delete")
    def update_cache_template(sender, instance, **kwargs):
        """
        Clear the cache for content blocks when their db template is saved.
        :param instance:
        :param sender:
        :param kwargs:
        :return:
        """
        if kwargs.get("raw", False):
            # Prevent this signal from running during loaddata.
            return

        content_blocks = ContentBlock.objects.filter(
            content_block_template__template_filename=instance.name.split("/")[-1]
        )

        for content_block in content_blocks:
            content_block.update_cache()


def cleanup_media(sender, instance, delete=False, **kwargs):
    """
    Delete old media files when no longer needed.
    :param sender:
    :param instance:
    :param delete:
    :param kwargs:
    :return:
    """
    if kwargs.get("raw", False):
        # Prevent this signal from running during loaddata.
        return

    if instance.id is None:
        return  # New object being created

    object_types = {
        ContentBlockFields.IMAGE_FIELD: "image",
        ContentBlockFields.FILE_FIELD: "file",
        ContentBlockFields.VIDEO_FIELD: "video",
    }

    object_type = object_types.get(instance.field_type)
    if object_type is None:
        return  # pragma: no cover

    if delete:
        old_file = getattr(instance, object_type)
        new_file = None
    else:
        try:
            old_file = getattr(
                ContentBlockField.objects.get(id=instance.id), object_type
            )
            new_file = getattr(instance, object_type)
        except ContentBlockField.DoesNotExist:  # pragma: no cover
            return

    if (
        old_file != new_file
        and not ContentBlockField.objects.filter(**{object_type: old_file})
        .exclude(id=instance.id)
        .exists()
    ):
        old_file.delete(save=False)


@receiver(pre_save, sender=ImageField, dispatch_uid="cleanup_image_media_save")
@receiver(pre_save, sender=FileField, dispatch_uid="cleanup_file_media_save")
@receiver(pre_save, sender=VideoField, dispatch_uid="cleanup_video_media_save")
def cleanup_media_save(sender, instance, **kwargs):
    return cleanup_media(sender, instance, delete=False, **kwargs)


@receiver(pre_delete, sender=ImageField, dispatch_uid="cleanup_image_media_delete")
@receiver(pre_delete, sender=FileField, dispatch_uid="cleanup_file_media_delete")
@receiver(pre_delete, sender=VideoField, dispatch_uid="cleanup_video_media_delete")
def cleanup_media_delete(sender, instance, **kwargs):
    return cleanup_media(sender, instance, delete=True, **kwargs)



# Generated by Django 4.1.7 on 2023-03-17 11:59

import content_blocks.fields
import content_blocks.models
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import model_clone.mixin


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='ContentBlock',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('visible', models.BooleanField(default=True, help_text='Uncheck to hide this.')),
                ('create_date', models.DateTimeField(auto_now_add=True, verbose_name='Creation Date')),
                ('mod_date', models.DateTimeField(auto_now=True, verbose_name='Last Modified')),
                ('position', models.PositiveIntegerField(default=0, help_text='Set a custom ordering. Lower numbers appear first.')),
                ('css_class', models.CharField(blank=True, max_length=256)),
                ('draft', models.BooleanField(blank=True, default=False)),
                ('saved', models.BooleanField(blank=True, default=False)),
            ],
            options={
                'ordering': ['position'],
                'abstract': False,
            },
            bases=(model_clone.mixin.CloneMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ContentBlockTemplate',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('visible', models.BooleanField(default=True, help_text='Uncheck to hide this.')),
                ('create_date', models.DateTimeField(auto_now_add=True, verbose_name='Creation Date')),
                ('mod_date', models.DateTimeField(auto_now=True, verbose_name='Last Modified')),
                ('position', models.PositiveIntegerField(default=0, help_text='Set a custom ordering. Lower numbers appear first.')),
                ('name', models.CharField(max_length=256, unique=True)),
                ('template_filename', models.CharField(blank=True, max_length=256)),
                ('no_cache', models.BooleanField(default=False, help_text='Disable caching for content blocks created with this template.')),
            ],
            options={
                'ordering': ['position'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ContentBlockTemplateField',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('position', models.PositiveIntegerField(default=0, help_text='Set a custom ordering. Lower numbers appear first.')),
                ('field_type', models.CharField(choices=[('TextField', 'Text Field'), ('ContentField', 'Content Field'), ('ImageField', 'Image Field'), ('VideoField', 'Video Field'), ('FileField', 'File Field'), ('EmbeddedVideoField', 'Embedded Video Field'), ('NestedField', 'Nested Field'), ('ModelChoiceField', 'Model Choice Field'), ('ChoiceField', 'Choice Field'), ('CheckboxField', 'Checkbox Field')], max_length=256)),
                ('key', models.SlugField(help_text='Must be unique to this content block template. Lowercase letters, numbers and underscores only.', max_length=256, validators=[django.core.validators.RegexValidator('[a-z0-9_]+', 'Lowercase letters, numbers and underscores only.')])),
                ('help_text', models.TextField(blank=True)),
                ('required', models.BooleanField(blank=True, default=False)),
                ('css_class', models.CharField(blank=True, help_text='Set a custom CSS class for this field in the editor.', max_length=256)),
                ('min_num', models.PositiveIntegerField(default=0, help_text='The minimum number of nested blocks allowed.')),
                ('max_num', models.PositiveIntegerField(default=99, help_text='The maximum number of nested blocks allowed.')),
                ('choices', models.TextField(blank=True)),
                ('content_block_template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='content_block_template_fields', to='content_blocks.contentblocktemplate')),
                ('model_choice_content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('nested_templates', models.ManyToManyField(blank=True, help_text='Choose the content block templates that can be used in this nested field.', to='content_blocks.contentblocktemplate')),
            ],
            options={
                'ordering': ['position'],
                'abstract': False,
                'unique_together': {('key', 'content_block_template')},
            },
        ),
        migrations.CreateModel(
            name='ContentBlockField',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('field_type', models.CharField(max_length=256)),
                ('text', models.CharField(blank=True, max_length=256)),
                ('content', models.TextField(blank=True)),
                ('checkbox', models.BooleanField(blank=True, default=False)),
                ('image', content_blocks.fields.SVGAndImageField(blank=True, storage=content_blocks.models.image_storage, upload_to='content-blocks/images')),
                ('file', models.FileField(blank=True, storage=content_blocks.models.file_storage, upload_to='content-blocks/files')),
                ('choice', models.CharField(blank=True, max_length=256)),
                ('video', content_blocks.fields.VideoField(blank=True, storage=content_blocks.models.video_storage, upload_to='content-blocks/videos')),
                ('embedded_video', models.CharField(blank=True, max_length=256)),
                ('model_choice_object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('content_block', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='content_block_fields', to='content_blocks.contentblock')),
                ('model_choice_content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='contenttypes.contenttype')),
                ('template_field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='fields', to='content_blocks.contentblocktemplatefield')),
            ],
            options={
                'ordering': ['template_field__position'],
            },
            bases=(models.Model, model_clone.mixin.CloneMixin),
        ),
        migrations.CreateModel(
            name='ContentBlockCollection',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('create_date', models.DateTimeField(auto_now_add=True, verbose_name='Creation Date')),
                ('mod_date', models.DateTimeField(auto_now=True, verbose_name='Last Modified')),
                ('name', models.CharField(blank=True, help_text='For identification purposes only.', max_length=256)),
                ('slug', models.SlugField(unique=True)),
                ('content_blocks', models.ManyToManyField(to='content_blocks.contentblock')),
            ],
            options={
                'ordering': ['-create_date'],
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ContentBlockAvailability',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('create_date', models.DateTimeField(auto_now_add=True, verbose_name='Creation Date')),
                ('mod_date', models.DateTimeField(auto_now=True, verbose_name='Last Modified')),
                ('content_block_templates', models.ManyToManyField(to='content_blocks.contentblocktemplate')),
                ('content_type', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
            ],
            options={
                'verbose_name_plural': 'Content block availabilities',
            },
        ),
        migrations.AddField(
            model_name='contentblock',
            name='content_block_template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='content_blocks.contentblocktemplate'),
        ),
        migrations.AddField(
            model_name='contentblock',
            name='parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='content_blocks', to='content_blocks.contentblockfield'),
        ),
        migrations.CreateModel(
            name='CheckboxField',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('content_blocks.contentblockfield',),
        ),
        migrations.CreateModel(
            name='ChoiceField',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('content_blocks.contentblockfield',),
        ),
        migrations.CreateModel(
            name='ContentField',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('content_blocks.contentblockfield',),
        ),
        migrations.CreateModel(
            name='EmbeddedVideoField',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('content_blocks.contentblockfield',),
        ),
        migrations.CreateModel(
            name='FileField',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('content_blocks.contentblockfield',),
        ),
        migrations.CreateModel(
            name='ImageField',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('content_blocks.contentblockfield',),
        ),
        migrations.CreateModel(
            name='ModelChoiceField',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('content_blocks.contentblockfield',),
        ),
        migrations.CreateModel(
            name='NestedField',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('content_blocks.contentblockfield',),
        ),
        migrations.CreateModel(
            name='TextField',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('content_blocks.contentblockfield',),
        ),
        migrations.CreateModel(
            name='VideoField',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('content_blocks.contentblockfield',),
        ),
    ]



from django import forms
from django.contrib.contenttypes.models import ContentType
from django.db import transaction

from content_blocks.models import (
    ContentBlock,
    ContentBlockAvailability,
    ContentBlockField,
    ContentBlockFields,
    ContentBlockTemplate,
)


class ParentModelForm(forms.Form):
    """
    Adds parent object to form from kwargs.
    """

    parent = None

    def __init__(self, *args, **kwargs):
        self.parent = kwargs.pop("parent")
        super().__init__(*args, **kwargs)


class NewContentBlockFormBase(forms.Form):
    """
    Shared functionality between NewContentBlockForm and NewNestedBlockForm.
    Has a select for content block template.
    """

    content_block_template = forms.ModelChoiceField(
        ContentBlockTemplate.objects.visible(), empty_label=None
    )

    position = forms.CharField(initial=0, widget=forms.HiddenInput())

    auto_id = True

    def __init__(self, *args, **kwargs):
        kwargs["auto_id"] = self.auto_id
        super().__init__(*args, **kwargs)
        self.fields["content_block_template"].queryset = self.get_available_templates()

    def get_available_templates(self):
        """
        Must be provided by subclasses.
        """
        raise NotImplementedError  # pragma: no cover

    def create_content_block(self, content_block_template, draft=False, **kwargs):
        content_block = ContentBlock.objects.create(
            content_block_template=content_block_template,
            draft=draft,
            position=self.cleaned_data.get("position", 0),
            **kwargs,
        )

        for (
            template_field
        ) in content_block_template.content_block_template_fields.all():
            content_block_field = ContentBlockField.objects.create(
                content_block=content_block,
                template_field=template_field,
                field_type=template_field.field_type,
                model_choice_content_type=template_field.model_choice_content_type,
            )

            # create min_num nested blocks
            if template_field.field_type == ContentBlockFields.NESTED_FIELD:
                for j in range(template_field.min_num):
                    self.create_content_block(
                        template_field.nested_templates.first(),
                        draft=draft,
                        parent=content_block_field,
                    )

        return content_block


class NewContentBlockForm(ParentModelForm, NewContentBlockFormBase):
    """
    Form for creating new top level content blocks.
    Creates new content block object with appropriate fields on save.
    Updates the parent which the content block belongs to m2m on save.
    """

    auto_id = "new_cb_%s"

    def get_available_templates(self):
        try:
            content_type = ContentType.objects.get_for_model(self.parent)
            content_block_availability = ContentBlockAvailability.objects.get(
                content_type=content_type
            )
            return content_block_availability.content_block_templates.visible()
        except ContentBlockAvailability.DoesNotExist:
            return ContentBlockTemplate.objects.visible()

    def update_parent_m2m(self, content_block):
        self.parent.content_blocks.add(content_block)

    def save(self):
        content_block = self.create_content_block(
            self.cleaned_data["content_block_template"], draft=True
        )
        self.update_parent_m2m(content_block)
        return content_block


class NewNestedBlockForm(NewContentBlockFormBase):
    """
    As per NewContentBlockForm but for adding nested blocks.
    """

    parent = forms.ModelChoiceField(
        widget=forms.HiddenInput(), queryset=ContentBlockField.objects.all()
    )

    auto_id = False

    def get_available_templates(self):
        try:
            parent = self.initial["parent"]
            return parent.template_field.nested_templates.visible()
        except KeyError:
            return ContentBlockTemplate.objects.all()

    def save(self):
        return self.create_content_block(
            self.cleaned_data["content_block_template"],
            draft=False,
            parent=self.cleaned_data["parent"],
        )


class ContentBlockForm(forms.Form):
    """
    Form for creating and changing content blocks.
    Form fields are generated from the content block fields.
    On save each content block field is updated and saved.
    """

    css_class = forms.CharField(
        max_length=256,
        required=False,
        label="CSS Class",
        help_text="For advanced styling.",
    )

    def __init__(self, *args, **kwargs):
        self.content_block = kwargs.pop("content_block")
        kwargs["auto_id"] = f"id_%s_{self.content_block.id}"
        super().__init__(*args, **kwargs)
        self.set_fields()
        self.fields["css_class"].initial = self.content_block.css_class

    def set_fields(self):
        for key, field in self.content_block.fields.items():
            if field.form_field:
                self.fields[key] = field.form_field

    def save(self):
        with transaction.atomic():
            for key, field in self.content_block.fields.items():
                if key in self.cleaned_data.keys():
                    field.save_value(self.cleaned_data.get(key))

            self.content_block.css_class = self.cleaned_data.get("css_class")
            self.content_block.saved = True
            self.content_block.save()
            # This is problematic as parent blocks can try to render without nested blocks validating
            # self.content_block.update_cache()
            self.content_block.clear_cache()
        return self.content_block


class PublishContentBlocksForm(ParentModelForm):
    """
    Form for publishing content blocks.
    Duplicates all content blocks and fields and nested blocks and sets draft=False
    """

    def save(self):
        self.parent.content_blocks.published().delete()

        for content_block in self.parent.content_blocks.drafts():
            new_content_block = content_block.clone(attrs={"draft": False})
            self.parent.content_blocks.add(new_content_block)
            # Cache the html
            new_content_block.render()


class ResetContentBlocksForm(ParentModelForm):
    """
    Deletes existing draft content blocks and creates new from published content blocks.
    """

    def save(self):
        self.parent.content_blocks.drafts().delete()

        for content_block in self.parent.content_blocks.published():
            new_content_block = content_block.clone(attrs={"draft": True})
            self.parent.content_blocks.add(new_content_block)


class ImportContentBlocksForm(ParentModelForm):
    """
    Import content blocks from another master object of the same type as parent.
    """

    master = forms.ModelChoiceField(None, empty_label=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["master"].queryset = self.get_master_queryset()

    def get_master_queryset(self):
        return self.parent._meta.model.objects.exclude(id=self.parent.id)

    def save(self):
        self.parent.content_blocks.drafts().delete()

        for content_block in self.cleaned_data["master"].content_blocks.drafts():
            new_content_block = content_block.clone()
            self.parent.content_blocks.add(new_content_block)



"""
Content Blocks views.py
"""
from django.contrib.admin.models import ADDITION, CHANGE, DELETION, LogEntry
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.contenttypes.models import ContentType
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, render
from django.template.loader import render_to_string
from django.urls import reverse
from django.views.decorators.csrf import ensure_csrf_cookie
from django.views.decorators.http import require_POST

from content_blocks.conf import settings
from content_blocks.forms import (
    ContentBlockForm,
    ImportContentBlocksForm,
    NewContentBlockForm,
    NewNestedBlockForm,
    PublishContentBlocksForm,
    ResetContentBlocksForm,
)
from content_blocks.models import ContentBlock
from content_blocks.permissions import require_ajax


def create_log_entry(request, obj, action_flag, change_message, **kwargs):
    params = {
        "user_id": request.user.id,
        "content_type_id": ContentType.objects.get_for_model(obj.__class__).id,
        "object_id": obj.id,
        "object_repr": obj.__str__(),
        "action_flag": action_flag,
        "change_message": change_message,
    }

    params.update(kwargs)

    LogEntry.objects.log_action(**params)


def render_with_admin_context(request, template, context, model_admin):
    context.update(**model_admin.admin_site.each_context(request))
    return render(request, template, context)


@staff_member_required
@ensure_csrf_cookie
def content_block_editor(request, object_id, model_admin=None):
    """
    Show the content block editor.
    """
    parent = get_object_or_404(model_admin.model, id=object_id)

    app_label, model_name = parent._meta.app_label, parent._meta.model_name

    return_url = reverse(f"admin:{app_label}_{model_name}_change", args=[parent.id])

    new_content_block_form = NewContentBlockForm(parent=parent)
    publish_form = PublishContentBlocksForm(
        initial={
            "return_url": return_url,
        },
        parent=parent,
    )
    discard_changes_form = ResetContentBlocksForm(
        initial={
            "return_url": return_url,
        },
        parent=parent,
    )
    import_content_blocks_form = ImportContentBlocksForm(parent=parent)

    status_message = settings.CONTENT_BLOCKS_DEFAULT_STATUS_MESSAGE
    if callable(status_message):
        status_message = status_message()

    return render_with_admin_context(
        request,
        "content_blocks/editor/editor.html",
        {
            "import_content_blocks_form": import_content_blocks_form,
            "new_content_block_form": new_content_block_form,
            "parent": parent,
            "publish_form": publish_form,
            "discard_form": discard_changes_form,
            "return_url": return_url,
            "opts": parent._meta,
            "status_message": status_message,
        },
        model_admin,
    )


def content_block_create_base(request, form, parent, log_name="content block"):
    """
    Ajax view for creating new top level and nested content blocks depending on the form passed.
    Returns html form for the new content block.
    """
    # This should never happen
    if not form.is_valid():
        return JsonResponse({"error": f"Invalid {form.__class__}"})  # pragma: no cover

    content_block = form.save()

    create_log_entry(
        request,
        content_block,
        ADDITION,
        [{"added": {"name": log_name, "object": content_block.__str__()}}],
    )

    content_block_form_html = render_to_string(
        "content_blocks/editor/content_block_form_wrapper.html",
        {
            "form": ContentBlockForm(content_block=content_block),
            "content_block": content_block,
            "opts": parent._meta,
            "parent": parent,
        },
    )

    return JsonResponse({"html": content_block_form_html})


@staff_member_required
@require_POST
@require_ajax
def content_block_create(request, object_id, model_admin=None):
    parent = get_object_or_404(model_admin.model, id=object_id)
    return content_block_create_base(
        request,
        NewContentBlockForm(request.POST, parent=parent),
        parent,
    )


@staff_member_required
@require_POST
@require_ajax
def nested_block_create(request, object_id, model_admin=None):
    parent = get_object_or_404(model_admin.model, id=object_id)
    return content_block_create_base(
        request,
        NewNestedBlockForm(request.POST),
        parent,
        log_name="nested block",
    )


@staff_member_required
@require_POST
@require_ajax
def content_block_save(request, content_block_id):
    """
    Save the content block.  Return form html.
    """
    content_block = get_object_or_404(ContentBlock, id=content_block_id)

    form = ContentBlockForm(request.POST, request.FILES, content_block=content_block)

    form_is_valid = form.is_valid()

    if form_is_valid:
        form.save()

        create_log_entry(
            request, content_block, CHANGE, [{"changed": {"fields": form.changed_data}}]
        )

    content_block_form_html = render_to_string(
        "content_blocks/editor/content_block_form.html",
        {
            "form": ContentBlockForm(content_block=content_block)
            if form_is_valid
            else form,
            "content_block": content_block,
            "saved": form_is_valid,
        },
    )

    return JsonResponse({"html": content_block_form_html, "saved": form_is_valid})


@staff_member_required
@require_POST
@require_ajax
def update_position(request):
    """
    Update position after a content block has been dragon dropped.
    """
    positions = request.POST.get("positions", "")
    positions = f"&{positions}".split("&cb[]=")[1:]
    for i, id in enumerate(positions):
        ContentBlock.objects.filter(id=id).update(position=i)
    # Admin log entry?  Might be a bit spammy. Unless we pass the cb which was dragged and just log on that one?
    return JsonResponse({})


@staff_member_required
@require_POST
@require_ajax
def toggle_visible(request, content_block_id):
    """
    Toggle visible buttons
    """
    content_block = ContentBlock.objects.get(id=content_block_id)
    content_block.visible = not content_block.visible
    content_block.save()

    create_log_entry(
        request,
        content_block,
        action_flag=CHANGE,
        change_message=[{"changed": {"fields": ["visible"]}}],
    )

    return JsonResponse({"visible": content_block.visible})


def ajax_form(
    request,
    object_id,
    form_klass,
    content_blocks_html=False,
    model_admin=None,
):
    """
    Used by import, reset and publish to process a form and optionally return the parent's content blocks html.
    :param model_admin:
    :param object_id:
    :param content_blocks_html:
    :param form_klass:
    :param request:
    :return:
    """
    parent = get_object_or_404(model_admin.model, id=object_id)
    form = form_klass(request.POST, parent=parent)

    # This should never happen
    if not form.is_valid():
        return JsonResponse(
            {"error": f"Invalid {form_klass.__class__}"}
        )  # pragma: no cover

    form.save()

    data = {}
    if content_blocks_html:
        data["html"] = render_to_string(
            "content_blocks/editor/content_block_forms.html",
            {"parent": form.parent, "opts": parent._meta},
        )

    log_entry_name = {
        PublishContentBlocksForm: "published",
        ResetContentBlocksForm: "reset",
        ImportContentBlocksForm: "imported",
    }

    create_log_entry(
        request, parent, CHANGE, f"Content blocks {log_entry_name[form_klass]}"
    )

    return JsonResponse(data)


@staff_member_required
@require_POST
@require_ajax
def publish_content_blocks(request, object_id, model_admin=None):
    return ajax_form(
        request, object_id, PublishContentBlocksForm, model_admin=model_admin
    )


@staff_member_required
@require_POST
@require_ajax
def discard_changes(request, object_id, model_admin=None):
    return ajax_form(
        request,
        object_id,
        ResetContentBlocksForm,
        content_blocks_html=True,
        model_admin=model_admin,
    )


@staff_member_required
@require_POST
@require_ajax
def import_content_blocks(request, object_id, model_admin=None):
    return ajax_form(
        request,
        object_id,
        ImportContentBlocksForm,
        content_blocks_html=True,
        model_admin=model_admin,
    )


@staff_member_required
def content_block_preview(request, object_id, content_block_id, model_admin=None):
    """
    Render a content block in a preview base template and return html.
    Normally shown in an iframe.
    """
    parent = get_object_or_404(model_admin.model, id=object_id)
    content_block = get_object_or_404(ContentBlock, id=content_block_id)

    response = render(
        request,
        "content_blocks/content_block_preview.html",
        {"content_block": content_block, "parent": parent},
    )
    response.headers["X-Frame-Options"] = "SAMEORIGIN"

    return response


@staff_member_required
@require_POST
@require_ajax
def content_block_delete(request, content_block_id):
    """
    Delete a content block.
    """
    content_block = ContentBlock.objects.get(id=content_block_id)

    create_log_entry(request, content_block, DELETION, "")

    content_block.delete()
    return JsonResponse({})



"""
Content blocks admin.py
"""
from adminsortable2.admin import SortableAdminMixin, SortableInlineAdminMixin
from django.conf import settings
from django.contrib import admin, messages
from django.http import HttpResponseRedirect
from django.template import TemplateDoesNotExist
from django.template.loader import get_template
from django.urls import path, reverse
from django.utils.safestring import mark_safe

from content_blocks.admin_forms import (
    ContentBlockTemplateAdminForm,
    ContentBlockTemplateFieldAdminForm,
)
from content_blocks.models import (
    ContentBlockAvailability,
    ContentBlockCollection,
    ContentBlockTemplate,
    ContentBlockTemplateField,
)
from content_blocks.views import (
    content_block_create,
    content_block_editor,
    content_block_preview,
    discard_changes,
    import_content_blocks,
    nested_block_create,
    publish_content_blocks,
)

"""
Admin constants and base classes
"""

# Auto dates
AUTO_DATE_FIELDS = ["create_date", "mod_date"]
AUTO_DATE_FIELDSET = (
    "Dates",
    {"fields": (("create_date", "mod_date"),)},
)

# Visible
VISIBLE_FIELDS = ["visible"]
VISIBLE_FIELDSET = ("Visibility", {"fields": VISIBLE_FIELDS})


class ContentBlockTemplateFieldInline(SortableInlineAdminMixin, admin.StackedInline):
    model = ContentBlockTemplateField
    fk_name = "content_block_template"
    form = ContentBlockTemplateFieldAdminForm
    min_num = 0
    extra = 0
    filter_horizontal = ["nested_templates"]


@admin.register(ContentBlockTemplate)
class ContentBlockTemplateAdmin(SortableAdminMixin, admin.ModelAdmin):
    form = ContentBlockTemplateAdminForm

    list_display = [
        "name",
        "template_filename",
        "visible",
    ] + AUTO_DATE_FIELDS
    list_editable = ("visible",)
    list_filter = AUTO_DATE_FIELDS + ["visible"]
    search_fields = ("name", "template_filename")

    readonly_fields = AUTO_DATE_FIELDS
    inlines = [ContentBlockTemplateFieldInline]

    class Media:
        css = {"all": ["content_blocks/css/content_block_template_admin.css"]}
        js = (
            "content_blocks/js/content_block_template_admin.js",
            "content_blocks/js/admin_choices_widget.js",
        )

    @property
    def content_block_template_fields(self):
        return ["name", "template_filename", "no_cache"]

    def get_fieldsets(self, request, obj=None):
        return (
            AUTO_DATE_FIELDSET,
            VISIBLE_FIELDSET,
            (
                "Content Block Template",
                {
                    "fields": self.content_block_template_fields,
                },
            ),
        )

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)

        template_filename = form.cleaned_data.get("template_filename")

        if not template_filename:
            return

        template_filename = f"content_blocks/content_blocks/{template_filename}"

        try:
            get_template(template_filename)
        except TemplateDoesNotExist:
            url = reverse(
                "admin:content_blocks_contentblocktemplate_change", args=[obj.id]
            )
            msg = (
                f"Couldn't load a template for the content block template "
                f'"<a href="{url}">{obj}</a>". '
                f'No template found at "{template_filename}".'
            )
            messages.warning(request, mark_safe(msg))


if "dbtemplates" in settings.INSTALLED_APPS:
    from dbtemplates.models import Template

    admin.site.unregister(ContentBlockTemplate)

    @admin.register(ContentBlockTemplate)
    class ContentBlockTemplateDBTemplatesAdmin(ContentBlockTemplateAdmin):
        @property
        def content_block_template_fields(self):
            return super().content_block_template_fields + ["db_template_button"]

        def db_template_button(self, obj):
            if not obj.template_filename:
                return "-"
            button = '<input type="submit" value="Save and edit HTML template" name="_dbtemplate">'
            return mark_safe(button)

        db_template_button.short_description = "HTML Template"

        def get_readonly_fields(self, request, obj=None):
            return [
                *super().get_readonly_fields(request, obj=obj),
                "db_template_button",
            ]

        def _response(self, response, request, obj):
            if "_dbtemplate" in request.POST and obj.template:
                db_template, _ = Template.objects.get_or_create(name=obj.template)
                url = reverse(
                    "admin:dbtemplates_template_change", args=[db_template.id]
                )
                return HttpResponseRedirect(url)

            return response

        def response_change(self, request, obj):
            response = super().response_change(request, obj)
            return self._response(response, request, obj)

        def response_add(self, request, obj, post_url_continue=None):
            response = super().response_add(
                request, obj, post_url_continue=post_url_continue
            )
            return self._response(response, request, obj)


@admin.register(ContentBlockAvailability)
class ContentBlockAvailabilityAdmin(admin.ModelAdmin):
    list_display = ["content_type"] + AUTO_DATE_FIELDS
    list_filter = AUTO_DATE_FIELDS

    readonly_fields = AUTO_DATE_FIELDS
    filter_horizontal = ["content_block_templates"]

    fieldsets = (
        AUTO_DATE_FIELDSET,
        (
            "Content Block Availability",
            {"fields": ("content_type", "content_block_templates")},
        ),
    )


class ContentBlockModelAdmin(admin.ModelAdmin):
    """
    Base class to be added to the admin of any model which has a content_blocks m2m.  This will then add
    content_blocks_button which can then be added to the formset.
    """

    def content_blocks_link(self, obj):
        """
        A link to the content block editor for this object.
        Intended to be used on admin changelist.
        :param obj: subclass of :py:class:`ContentBlockParentModel`.
        """
        app_label, model_name = self.model._meta.app_label, self.model._meta.model_name
        url = reverse(
            f"admin:{app_label}_{model_name}_content_block_editor", args=[obj.id]
        )
        link = f'<a href="{url}">Edit content blocks</a>'
        return mark_safe(link)

    content_blocks_link.short_description = "Content blocks"

    def content_blocks_button(self, *args):
        """
        A button which will save the object then, if successful, will redirect to the content block editor for this
        object.
        Intended to be used on the admin changepage.
        """
        button = '<input type="submit" value="Save and edit content blocks" name="_contentblocks">'
        return mark_safe(button)

    content_blocks_button.short_description = "Content blocks"

    def get_readonly_fields(self, *args, **kwargs):
        return (
            *super().get_readonly_fields(*args, **kwargs),
            "content_blocks_link",
            "content_blocks_button",
        )

    def _response(self, response, request, obj):
        if "_contentblocks" in request.POST:
            model_name = self.model._meta.model_name
            app_label = self.model._meta.app_label
            url = reverse(
                f"admin:{app_label}_{model_name}_content_block_editor", args=[obj.id]
            )
            return HttpResponseRedirect(url)

        return response

    def response_change(self, request, obj):
        response = super().response_change(request, obj)
        return self._response(response, request, obj)

    def response_add(self, request, obj, post_url_continue=None):
        response = super().response_add(
            request, obj, post_url_continue=post_url_continue
        )
        return self._response(response, request, obj)

    def get_urls(self):
        app_label, model_name = self.model._meta.app_label, self.model._meta.model_name
        urls = [
            path(
                "<path:object_id>/content-blocks/",
                content_block_editor,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_content_block_editor",
            ),
            path(
                "<path:object_id>/content-blocks/_ajax/content-block-create/",
                content_block_create,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_content_block_create",
            ),
            path(
                "<path:object_id>/content-blocks/_ajax/nested-block-create/",
                nested_block_create,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_nested_block_create",
            ),
            path(
                "<path:object_id>/content-blocks/_ajax/publish/",
                publish_content_blocks,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_publish_content_blocks",
            ),
            path(
                "<path:object_id>/content-blocks/_ajax/discard-changes/",
                discard_changes,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_discard_changes",
            ),
            path(
                "<path:object_id>/content-blocks/_ajax/import-content-blocks/",
                import_content_blocks,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_import_content_blocks",
            ),
            path(
                "<path:object_id>/content-blocks/<int:content_block_id>/preview/",
                content_block_preview,
                {"model_admin": self},
                name=f"{app_label}_{model_name}_content_block_preview",
            ),
        ]
        return urls + super().get_urls()


@admin.register(ContentBlockCollection)
class ContentBlockCollectionAdmin(ContentBlockModelAdmin):
    list_display = ["name", "slug", "content_blocks_link"] + AUTO_DATE_FIELDS
    list_display_links = ["name", "slug"]
    list_filter = AUTO_DATE_FIELDS
    search_fields = ["name", "slug"]

    readonly_fields = AUTO_DATE_FIELDS
    prepopulated_fields = {"slug": ("name",)}

    fieldsets = (
        AUTO_DATE_FIELDSET,
        (
            "Content Block Collection",
            {"fields": ("name", "slug", "content_blocks_button")},
        ),
    )



"""
Content Blocks urls.py
"""
from django.urls import path

from content_blocks.views import (
    content_block_delete,
    content_block_save,
    toggle_visible,
    update_position,
)

app_name = "content_blocks"
urlpatterns = [
    path(
        "_ajax/<int:content_block_id>/save/",
        content_block_save,
        name="content_block_save",
    ),
    path("_ajax/update-position/", update_position, name="update_position"),
    path(
        "_ajax/<int:content_block_id>/toggle-visible/",
        toggle_visible,
        name="toggle_visible",
    ),
    path(
        "_ajax/<int:content_block_id>/delete/",
        content_block_delete,
        name="content_block_delete",
    ),
]



"""
Content Blocks models.py
"""
import json
import logging

from django import forms
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.files.storage import get_storage_class
from django.core.validators import RegexValidator
from django.db import models
from django.template.loader import render_to_string
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from model_clone import CloneMixin

from content_blocks.abstract_models import (
    AutoDateModel,
    CachedHtmlModel,
    PositionModel,
    VisibleManager,
    VisibleModel,
)
from content_blocks.cache import cache
from content_blocks.conf import settings
from content_blocks.fields import (
    SVGAndImageField,
    SVGAndImageFieldFormField,
    VideoField,
)
from content_blocks.widgets import FileWidget

logger = logging.getLogger(__name__)

if (
    "django_cleanup" in settings.INSTALLED_APPS
    or "django_cleanup.apps.CleanupConfig" in settings.INSTALLED_APPS
):
    from django_cleanup.cleanup import cleanup_ignore
else:
    # noop shim for when django_cleanup isn't installed
    def cleanup_ignore(cls):  # pragma: no cover (covered in settings specific tests)
        return cls


class ContentBlockFields:
    # todo replace this with models.TextChoices
    TEXT_FIELD = "TextField"
    CONTENT_FIELD = "ContentField"
    IMAGE_FIELD = "ImageField"
    NESTED_FIELD = "NestedField"
    CHECKBOX_FIELD = "CheckboxField"
    CHOICE_FIELD = "ChoiceField"
    MODEL_CHOICE_FIELD = "ModelChoiceField"
    FILE_FIELD = "FileField"
    VIDEO_FIELD = "VideoField"
    EMBEDDED_VIDEO_FIELD = "EmbeddedVideoField"

    CHOICES = (
        (TEXT_FIELD, "Text Field"),
        (CONTENT_FIELD, "Content Field"),
        (IMAGE_FIELD, "Image Field"),
        (VIDEO_FIELD, "Video Field"),
        (FILE_FIELD, "File Field"),
        (EMBEDDED_VIDEO_FIELD, "Embedded Video Field"),
        (NESTED_FIELD, "Nested Field"),
        (MODEL_CHOICE_FIELD, "Model Choice Field"),
        (CHOICE_FIELD, "Choice Field"),
        (CHECKBOX_FIELD, "Checkbox Field"),
    )


class ContentBlockFieldManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().select_related("template_field")


def get_storage(field_type):
    """
    Get the storage class from dotted strings in settings.
    If the field storage class is in settings use that otherwise use the general storage class setting for content
    blocks, which will default to STORAGES["default"].
    :param field_type:
    :return:
    """
    storage_setting = getattr(settings, field_type, None)
    if storage_setting is not None:
        return get_storage_class(
            storage_setting
        )()  # pragma: no cover (covered by settings specific tests)
    return get_storage_class(settings.CONTENT_BLOCKS_STORAGE)()


def image_storage():
    return get_storage("CONTENT_BLOCKS_IMAGE_STORAGE")


def file_storage():
    return get_storage("CONTENT_BLOCKS_FILE_STORAGE")


def video_storage():
    return get_storage("CONTENT_BLOCKS_VIDEO_STORAGE")


@cleanup_ignore
class ContentBlockField(models.Model, CloneMixin):
    """
    Base model for all content block field models.
    """

    objects = ContentBlockFieldManager()
    template_field = models.ForeignKey(
        "content_blocks.ContentBlockTemplateField",
        on_delete=models.CASCADE,
        related_name="fields",
    )

    content_block = models.ForeignKey(
        "content_blocks.ContentBlock",
        on_delete=models.CASCADE,
        related_name="content_block_fields",
    )

    # duplicate from template_field here because it can't be changed on template field and seems impossible to have
    # select_related working in the __init__
    # todo fix the maxlength here 32 is more like it
    field_type = models.CharField(max_length=256)

    text = models.CharField(max_length=256, blank=True)
    content = models.TextField(blank=True)
    checkbox = models.BooleanField(blank=True, default=False)
    image = SVGAndImageField(
        upload_to="content-blocks/images", blank=True, storage=image_storage
    )
    file = models.FileField(
        upload_to="content-blocks/files", blank=True, storage=file_storage
    )
    choice = models.CharField(max_length=256, blank=True)
    video = VideoField(
        upload_to="content-blocks/videos", blank=True, storage=video_storage
    )
    embedded_video = models.CharField(max_length=256, blank=True)

    model_choice_content_type = models.ForeignKey(
        ContentType, on_delete=models.SET_NULL, blank=True, null=True
    )
    model_choice_object_id = models.PositiveIntegerField(blank=True, null=True)
    model_choice = GenericForeignKey(
        "model_choice_content_type", "model_choice_object_id"
    )

    template_name = "content_blocks/partials/fields/default.html"
    preview_template_name = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # todo remove this try
        try:
            for _class in ContentBlockField.__subclasses__():
                if self.field_type == _class.__name__:
                    self.__class__ = _class
                    break
        except:  # pragma: no cover # noqa
            logger.error("Polymorphism failed")  # pragma: no cover

    class Meta:
        ordering = ["template_field__position"]

    @property
    def form_field(self):
        """
        To be overridden. Will return a form.Field class if the field is to be shown on the form.
        This method should return _form_field() with the desired forms.Field.
        """
        raise NotImplementedError  # pragma: no cover

    def _form_field(self, field):
        """
        Adds content_block_field attribute to field
        :return:
        """
        # todo can this be replace with a decorator?
        field.cb_field = self
        return field

    @property
    def context_value(self):
        raise NotImplementedError  # pragma: no cover

    def save_value(self, value):
        raise NotImplementedError  # pragma: no cover

    @property
    def key(self):
        return self.template_field.key


class TextField(ContentBlockField):
    class Meta:
        proxy = True

    @property
    def context_value(self):
        if settings.CONTENT_BLOCKS_MARK_SAFE:
            return mark_safe(self.text)
        return self.text

    def save_value(self, value):
        self.text = value
        self.save()
        return value

    @property
    def form_field(self):
        widget = (
            forms.Textarea
            if settings.CONTENT_BLOCKS_TEXTFIELD_TEXTAREA
            else forms.TextInput
        )
        # noinspection PyTypeChecker
        return self._form_field(
            forms.CharField(
                initial=self.text,
                required=self.template_field.required,
                help_text=self.template_field.help_text,
                widget=widget,
            )
        )


class ContentField(ContentBlockField):
    class Meta:
        proxy = True

    @property
    def context_value(self):
        if settings.CONTENT_BLOCKS_MARK_SAFE:
            return mark_safe(self.content)
        return self.content

    def save_value(self, value):
        self.content = value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return self._form_field(
            forms.CharField(
                initial=self.content,
                required=self.template_field.required,
                help_text=self.template_field.help_text,
                widget=forms.Textarea(),
            )
        )


class CheckboxField(ContentBlockField):
    template_name = "content_blocks/partials/fields/checkbox.html"

    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.checkbox

    def save_value(self, value):
        self.checkbox = value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return self._form_field(
            forms.BooleanField(
                initial=self.checkbox,
                required=False,
                help_text=self.template_field.help_text,
            )
        )


@cleanup_ignore
class ImageField(ContentBlockField):
    preview_template_name = "content_blocks/partials/fields/previews/image.html"

    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.image

    def save_value(self, value):
        self.image = None if value is False else value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return self._form_field(
            SVGAndImageFieldFormField(
                initial=self.image,
                required=self.template_field.required,
                help_text=self.template_field.help_text,
                widget=FileWidget(),
            )
        )


@cleanup_ignore
class FileField(ContentBlockField):
    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.file

    def save_value(self, value):
        self.file = None if value is False else value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return self._form_field(
            forms.FileField(
                initial=self.file,
                required=self.template_field.required,
                help_text=self.template_field.help_text,
                widget=FileWidget(),
            )
        )


@cleanup_ignore
class VideoField(ContentBlockField):
    preview_template_name = "content_blocks/partials/fields/previews/video.html"

    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.video

    def save_value(self, value):
        self.video = None if value is False else value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return self._form_field(
            forms.FileField(
                initial=self.video,
                required=self.template_field.required,
                help_text=self.template_field.help_text,
                widget=FileWidget(),
            )
        )


class EmbeddedVideoField(ContentBlockField):
    preview_template_name = (
        "content_blocks/partials/fields/previews/embedded_video.html"
    )

    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.embedded_video

    def save_value(self, value):
        self.embedded_video = value
        self.save()
        return value

    @property
    def form_field(self):
        # noinspection PyTypeChecker
        return self._form_field(
            forms.CharField(
                initial=self.embedded_video,
                required=self.template_field.required,
                help_text=self.template_field.help_text,
            )
        )


class ChoiceField(ContentBlockField):
    class Meta:
        proxy = True

    @property
    def context_value(self):
        return self.choice

    def save_value(self, value):
        self.choice = value
        self.save()
        return value

    @property
    def form_field(self):
        choices = [[None, "-" * 9]] + json.loads(self.template_field.choices)
        # noinspection PyTypeChecker
        return self._form_field(
            forms.CharField(
                initial=self.choice,
                required=self.template_field.required,
                help_text=self.template_field.help_text,
                widget=forms.Select(choices=choices),
            )
        )


class ModelChoiceField(ContentBlockField):
    class Meta:
        proxy = True

    @property
    def form_field(self):
        # self.model_choice_content_type must be set or this will raise an error.
        model = self.model_choice_content_type.model_class()
        queryset = model.objects.all()
        # noinspection PyTypeChecker
        return self._form_field(
            forms.ModelChoiceField(
                queryset,
                initial=self.model_choice,
                required=self.template_field.required,
                help_text=self.template_field.help_text,
            )
        )

    @property
    def context_value(self):
        return self.model_choice

    def save_value(self, value):
        # If value is None, and we set self.model_choice = None it clears model_choice_content_type - not ideal
        # So instead we ste the model_choice_object_id
        self.model_choice_object_id = value.id if value is not None else None
        self.save()
        return value


class NestedField(ContentBlockField):
    class Meta:
        proxy = True

    @property
    def form_field(self):
        return None

    def save_value(self, value):
        return None

    @property
    def context_value(self):
        """
        :return: Nested content blocks which we can then call render or access context on.
        """
        return self.content_blocks.visible()


class ContentBlockManager(VisibleManager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .prefetch_related("content_block_fields")
            .select_related("content_block_template")
        )

    def visible(self):
        """
        Visible published only.
        Used in templates to render published content blocks.
        :return:
        """
        return super().visible().filter(draft=False)

    def previews(self):
        """
        Visible drafts only. Exclude those which haven't been saved via the editor yet (empties).
        Can be used in page previews.
        :return:
        """
        return super().visible().filter(draft=True, saved=True)

    def published(self):
        """
        All published including visible=False.
        Used by the publishing/reset mechanism.
        :return:
        """
        return self.get_queryset().filter(draft=False)

    def drafts(self):
        """
        All drafts including visible=False.
        Used by the editor and publishing/reset mechanism.
        :return:
        """
        return self.get_queryset().filter(draft=True)


class ContentBlock(
    PositionModel, AutoDateModel, VisibleModel, CloneMixin, CachedHtmlModel
):
    """
    Content Block Model
    Content block fields point here.  Those created are based on the content block template.
    """

    objects = ContentBlockManager()

    content_block_template = models.ForeignKey(
        "content_blocks.ContentBlockTemplate", on_delete=models.CASCADE
    )
    parent = models.ForeignKey(
        ContentBlockField,
        on_delete=models.CASCADE,
        related_name="content_blocks",
        blank=True,
        null=True,
    )

    # todo max_length=64
    css_class = models.CharField(max_length=256, blank=True)

    draft = models.BooleanField(blank=True, default=False)

    saved = models.BooleanField(blank=True, default=False)

    context_name = "content_block"
    cache_prefix = "content_block"

    @cached_property
    def template(self):
        """
        HTML template
        """
        return self.content_block_template.template or None

    @cached_property
    def fields(self):
        """
        Create/cache a dict of key to content block field objects.
        """
        nested_fields = {}
        fields = {}
        for field in self.content_block_fields.all():
            fields[field.template_field.key] = field
            if field.template_field.field_type == ContentBlockFields.NESTED_FIELD:
                nested_fields[field.template_field.key] = field

        self.__dict__["nested_fields"] = nested_fields
        return fields

    @cached_property
    def nested_fields(self):
        """
        Create/cache a dict of nested fields.  If self.fields has already been called this is already cached.
        """
        return {
            field.template_field.key: field
            for field in self.content_block_fields.filter(
                template_field__field_type=ContentBlockFields.NESTED_FIELD
            )
        }

    @cached_property
    def context(self):
        """
        Return dictionary of template context for this content block
        """
        return {key: field.context_value for key, field in self.fields.items()}

    def render(self):
        """
        Render html for this block and cache
        """
        if not self.can_render:
            return ""

        cache_enabled = (
            not self.content_block_template.no_cache
            and not settings.CONTENT_BLOCKS_DISABLE_CACHE
        )

        html = None
        if cache_enabled:
            html = cache.get(self.cache_key)

        if html is None:
            html = render_to_string(self.template, {self.context_name: self.context})

            if cache_enabled:
                cache.set(self.cache_key, html)

        return html

    def clear_cache(self):
        """
        Clear the cache for this block and any parent blocks
        """
        super().clear_cache()
        if self.parent:
            self.parent.content_block.clear_cache()

    def update_cache(self):
        if "context" in self.__dict__.keys():
            # Clear the context cached property if present
            del self.__dict__["context"]

        super().update_cache()
        if self.parent:
            self.parent.content_block.update_cache()

    def clone(self, attrs=None):
        """
        Clones the given content block and all content block fields.
        """
        new_content_block = self.make_clone(attrs=attrs)

        for field in self.content_block_fields.all():
            new_field = field.make_clone(attrs={"content_block": new_content_block})

            if field.template_field.field_type == ContentBlockFields.NESTED_FIELD:
                for nested_block in field.content_blocks.all():
                    nested_block.clone(attrs={"parent": new_field})

        return new_content_block


class ContentBlockTemplate(PositionModel, AutoDateModel, VisibleModel):
    """
    Content Block Template model.
    Used to define content block types.
    """

    objects = VisibleManager()

    name = models.CharField(max_length=256, unique=True)

    template_filename = models.CharField(max_length=256, blank=True)

    no_cache = models.BooleanField(
        default=False,
        help_text="Disable caching for content blocks created with this template.",
    )

    def __str__(self):
        return self.name

    @property
    def template(self):
        if self.template_filename:
            return f"content_blocks/content_blocks/{self.template_filename}"


class ContentBlockTemplateField(PositionModel):
    """
    Content Block Template Field model.
    Used to define content block types.
    """

    content_block_template = models.ForeignKey(
        ContentBlockTemplate,
        on_delete=models.CASCADE,
        related_name="content_block_template_fields",
    )

    # todo max_length=32
    field_type = models.CharField(max_length=256, choices=ContentBlockFields.CHOICES)

    # todo max_length=64
    key = models.SlugField(
        max_length=256,
        validators=[
            RegexValidator(
                "[a-z0-9_]+", "Lowercase letters, numbers and underscores only."
            )
        ],
        help_text="Must be unique to this content block template. Lowercase letters, numbers and underscores only.",
    )

    help_text = models.TextField(blank=True)
    required = models.BooleanField(blank=True, default=False)

    # todo max_length=64
    css_class = models.CharField(
        max_length=256,
        blank=True,
        help_text="Set a custom CSS class for this field in the editor.",
    )

    # Only used if field_type == NestedField
    nested_templates = models.ManyToManyField(
        "content_blocks.ContentBlockTemplate",
        blank=True,
        help_text="Choose the content block templates that can be used in this nested field.",
    )
    min_num = models.PositiveIntegerField(
        default=0, help_text="The minimum number of nested blocks allowed."
    )
    max_num = models.PositiveIntegerField(
        default=99, help_text="The maximum number of nested blocks allowed."
    )

    # Only used if field_type == ChoiceField
    choices = models.TextField(blank=True)

    # Only used if field_type == ModelChoiceField
    model_choice_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, blank=True, null=True
    )

    class Meta(PositionModel.Meta):
        unique_together = ("key", "content_block_template")

    def __str__(self):
        return self.key or super().__str__()


class ContentBlockAvailability(AutoDateModel):
    """
    Used to set which content block templates can be added to a model.
    """

    # null=True is required for admin list to work. blank=False because this is a required field.
    content_type = models.OneToOneField(
        ContentType, on_delete=models.CASCADE, null=True
    )

    content_block_templates = models.ManyToManyField(ContentBlockTemplate)

    class Meta:
        verbose_name_plural = "Content block availabilities"

    def __str__(self):
        return str(self.content_type)


class ContentBlockParentModel(models.Model):
    """
    Model to be inherited by parent objects which then have content blocks assigned.  E.g. imperium.Page
    """

    content_blocks = models.ManyToManyField(ContentBlock)

    class Meta:
        abstract = True

    def delete(self, using=None, keep_parents=False):
        self.content_blocks.all().delete()
        return super().delete(using=using, keep_parents=keep_parents)


class ContentBlockCollection(AutoDateModel, ContentBlockParentModel):
    """
    Simple collections of content blocks identified by slug.
    Use the content_blocks_collection template tag to render a content block collection in a template.
    """

    name = models.CharField(
        max_length=256, blank=True, help_text="For identification purposes only."
    )

    slug = models.SlugField(unique=True)

    def __str__(self):
        return self.name or self.slug



"""
Content Blocks permissions.py
"""
from functools import wraps

from django.core.exceptions import PermissionDenied


def require_ajax(view):
    @wraps(view)
    def _wrapped_view(request, *args, **kwargs):
        if request.META.get("HTTP_X_REQUESTED_WITH") == "XMLHttpRequest":
            return view(request, *args, **kwargs)
        else:
            raise PermissionDenied

    return _wrapped_view



from django.core.management import BaseCommand

from content_blocks.models import ContentBlock


class Command(BaseCommand):
    help = "Update content blocks cache"

    def handle(self, *args, **options):
        verbosity = int(options["verbosity"])

        for content_block in ContentBlock.objects.all():
            content_block.update_cache()

        if verbosity > 0:
            self.stdout.write("Content blocks cache updated")  # pragma: no cover



"""
Content Blocks conf.py
"""
from django.conf import settings as django_settings


class ContentBlocksSettings:
    # Choose which cache to use for content blocks
    CONTENT_BLOCKS_CACHE = "default"

    # Disable caching of content blocks
    CONTENT_BLOCKS_DISABLE_CACHE = False

    # Disable caching of content blocks on start
    CONTENT_BLOCKS_DISABLE_CACHE_ON_START = False

    # Disable updating cache on related model save
    CONTENT_BLOCKS_DISABLE_UPDATE_CACHE_MODEL_CHOICE = False

    # Choose the storage backend to use for media files
    # Image, field and video storage settings are use instead of content blocks storage if set.
    CONTENT_BLOCKS_IMAGE_STORAGE = None
    CONTENT_BLOCKS_FILE_STORAGE = None
    CONTENT_BLOCKS_VIDEO_STORAGE = None
    # Use default storage if not set.
    # Support provided here for Django 3.2
    CONTENT_BLOCKS_STORAGE = getattr(django_settings, "STORAGES", {}).get(
        "default", {}
    ).get("BACKEND") or getattr(
        django_settings,
        "DEFAULT_FILE_STORAGE",
        "django.core.files.storage.FileSystemStorage",
    )

    # Set the default status message.  Can be a callable.
    CONTENT_BLOCKS_DEFAULT_STATUS_MESSAGE = ""

    def __getattribute__(self, name):
        try:
            return getattr(django_settings, name)
        except AttributeError:
            return super().__getattribute__(name)

    # Mark context for TextField and ContentField as safe. Useful for allowing html in these fields.
    CONTENT_BLOCKS_MARK_SAFE = False

    # Use a textarea widget for TextFields.
    CONTENT_BLOCKS_TEXTFIELD_TEXTAREA = False


settings = ContentBlocksSettings()



"""
Content Blocks abstract_models.py
"""

from django.db import models
from django.template import TemplateDoesNotExist
from django.template.loader import get_template, render_to_string

from content_blocks.cache import cache


class VisibleManager(models.Manager):
    def visible(self):
        return self.get_queryset().filter(visible=True)


class VisibleModel(models.Model):
    visible = models.BooleanField(default=True, help_text="Uncheck to hide this.")

    class Meta:
        abstract = True


class AutoDateModel(models.Model):
    create_date = models.DateTimeField("Creation Date", auto_now_add=True)
    mod_date = models.DateTimeField("Last Modified", auto_now=True)

    class Meta:
        abstract = True
        ordering = ["-create_date"]


class PositionModel(models.Model):
    position = models.PositiveIntegerField(
        default=0, help_text="Set a custom ordering. Lower numbers appear first."
    )

    class Meta:
        abstract = True
        ordering = ["position"]


class CachedHtmlModel(models.Model):
    """
    Provides methods for cacheing objects html output and managing the cache.
    """

    _can_render = None
    context_name = "object"
    cache_prefix = None

    class Meta:
        abstract = True

    def delete(self, using=None, keep_parents=False):
        self.clear_cache()
        return super().delete(using=using, keep_parents=keep_parents)

    @property
    def template(self):
        """
        Must be overridden and return full path to template
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    @property
    def context(self):
        """
        Provided to be overridden with custom context for the object
        :return:
        """
        return self  # pragma: no cover

    @property
    def cache_key(self):
        """
        Set the cache prefix to something unique
        :return:
        """
        if not self.cache_prefix:
            raise NotImplementedError  # pragma: no cover

        return f"{self.cache_prefix}_{self.id}"

    @property
    def can_render(self):
        """
        Does the template exist
        :return:
        """
        if not self.template:
            return False
        try:
            get_template(self.template)
            return True
        except TemplateDoesNotExist:
            return False

    def render(self):  # pragma: no cover
        """
        Render html for this object and cache
        """
        if not self.can_render:
            return ""

        html = cache.get(self.cache_key)
        if not html:
            html = render_to_string(self.template, {self.context_name: self.context})
            cache.set(self.cache_key, html)
        return html

    def clear_cache(self):
        """
        Clear the cache for this object
        """
        cache.delete(self.cache_key)

    def update_cache(self):
        """
        Update the cache for this object
        :return:
        """
        cache.delete(self.cache_key)
        self.render()



import uuid
from pathlib import Path

from django import forms, template
from django.contrib.admin.widgets import AdminTextInputWidget
from django.utils.text import slugify


class FileWidget(forms.ClearableFileInput):
    template_name = "content_blocks/widgets/clearable_file.html"

    def clear_checkbox_id(self, name):
        return uuid.uuid4().hex


class TemplateFilenameAutocompleteWidget(AdminTextInputWidget):
    """
    Add autocomplete suggestions to text input via datalist.
    Suggestions are template file names in the given template_dir.
    """

    template_name = "content_blocks/widgets/filename_autocomplete.html"

    def __init__(self, template_dir=None, attrs=None):
        self.template_dir = template_dir or ""
        attrs = attrs or {}
        super().__init__(attrs={"list": self.data_list_id, **attrs})

    @property
    def template_list(self):
        template_dirs = []
        for engine in template.engines.all():
            template_dirs += [d for d in engine.template_dirs]

        file_names = []
        for template_dir in template_dirs:
            file_names += [
                p.name
                for p in Path(template_dir).glob(f"**{self.template_dir}/*.html")
                if p
            ]

        return sorted(file_names)

    @property
    def data_list_id(self):
        return f"{slugify(self.template_dir)}__suggestions"

    def get_context(self, *args):
        context = super().get_context(*args)
        context["widget"]["data_list_id"] = self.data_list_id
        context["widget"]["template_list"] = self.template_list
        return context
